import time
VERSION = 9974
STATUS_ONLINE = 0
STATUS_AWAY = 1

import binascii
import logging
import os

from lib.broadcast import Broadcast
from lib.direct import Direct
from state import State
from getdata import GetData
from message import STALE_PACKET, OUT_OF_ORDER_NET, OUT_OF_ORDER_SELF, OUT_OF_ORDER_BOTH
from message import DUPLICATE_PACKET
from message import MALFORMED_PACKET
from message import INVALID_SIGNATURE
from message import INVALID_HANDLE_ENCODING
from message import UNSUPPORTED_VERSION
from message import Message
from ignore import Ignore
from server import Server
from long_buffer import LongBuffer
from order_buffer import OrderBuffer
from short_buffer import ShortBuffer
from commands import BROADCAST, DIRECT, GETDATA, IGNORE

class Station(object):
    def __init__(self, cmd_line_options):
        self.client = None
        self.socket = None
        self.state = State(self, cmd_line_options.db_path)
        if cmd_line_options.address_table_path is not None:
            self.state.import_at_and_wot(cmd_line_options.address_table_path)
        self.short_buffer = ShortBuffer(self.state)
        self.long_buffer = LongBuffer(self.state)
        self.order_buffer = OrderBuffer(self.state)
        self.getdata_requests = {}
        self.server = Server(cmd_line_options, self)
        self.handlers = {
            DIRECT: self.handle_direct,
            BROADCAST: self.handle_broadcast,
            GETDATA: self.handle_getdata,
            IGNORE: self.handle_ignore
        }
        self.presence = {}
        self.last_delivered = 0

    def start(self):
        self.server.start()

    def handle_udp_data(self, bytes_address_pair):
        data = bytes_address_pair[0]
        address = bytes_address_pair[1]
        packet_info = (address[0],
                       address[1],
                       binascii.hexlify(data)[0:16])

        metadata = {
            'address': address,
            'packet_info': packet_info,
        }

        # if we don't have any peers let's just stop here
        if len(self.state.get_keyed_peers()) == 0:
            return

        for peer in self.state.get_keyed_peers():
            message = Message.unpack(peer,
                                     data,
                                     self.long_buffer,
                                     self.order_buffer,
                                     metadata,
                                     self.state)

            if message.get('error_code') is INVALID_SIGNATURE:
                continue

            if message.get('error_code') in [None,
                                             UNSUPPORTED_VERSION,
                                             STALE_PACKET,
                                             OUT_OF_ORDER_BOTH,
                                             OUT_OF_ORDER_SELF,
                                             OUT_OF_ORDER_NET,
                                             DUPLICATE_PACKET,
                                             MALFORMED_PACKET,
                                             INVALID_HANDLE_ENCODING]:
                break

        if message.get('error_code') is None:
            if message['command'] == DIRECT:
                self.handle_message(Direct(message, self.state))
            elif message['command'] == BROADCAST:
                self.handle_message(Broadcast(message, self.state))
            elif message['command'] == GETDATA:
                # This is a little weird.  We don't want to instantiate a GetData
                # object here because that would switch around body and self_chain,
                # so let's just instantiate a generic Message and make sure to handle
                # it as a GetData message.
                self.handle_message(Message(message, self.state))
            elif message['command'] == IGNORE:
                self.handle_message(Ignore(message, self.state))
        else:
            self.report_error(message['error_code'], message)

    def handle_message(self, message):
        try:
            self.handlers[message.command](message)
        except KeyError:
            logging.error("Unknown command, ignoring")

    def handle_direct(self, message):
        message.log_incoming(message.peer)
        self.deliver(message)
        self.conditionally_update_at(message, message.metadata["address"])

    def handle_broadcast(self, message):
        # it's possible we'll log dupes coming out of the order buffer here
        message.log_incoming(message.peer)

        # check if this is an immediate message
        if message.speaker in message.peer.handles:
            # remove message from short buffer if it was received as hearsay
            if self.short_buffer.has(message.message_hash):
                self.short_buffer.drop(message.message_hash)
            self.deliver(message)
            self.state.update_net_chain(message.message_hash)
            self.rebroadcast(message)
        else:
            # embargo to wait for immediate copy of message
            self.short_buffer.embargo(message)
        self.conditionally_update_at(message, message.metadata["address"])

    def handle_getdata(self, message):
        message.log_incoming_getdata(message.peer)

        # check for the requested message
        archived_message = self.long_buffer.exhume(message.body)

        # resend it if it exists
        if archived_message:
            archived_message.retry(message.peer)

    def handle_ignore(self, message):
        self.conditionally_update_at(message, message.metadata['address'])
        packet_info = message.metadata["packet_info"]
        address = packet_info[0]
        port = packet_info[1]
        packet_sample = packet_info[2]
        if os.environ.get('LOG_RUBBISH'):
            logging.debug("[%s:%d] -> ignoring packet: %s" % (address, port, packet_sample))
        return

    def check_order_buffer(self):
        messages = self.order_buffer.dequeue_and_order_mature_messages()
        for message in messages:
            self.handle_message(message)

    def check_short_buffer(self):
        messages = self.short_buffer.flush()
        for message_with_stats in messages:
            message = message_with_stats['message']
            # compute prefix
            if len(message_with_stats['closest_peers']) < 4:
                handles = []
                for peer in message_with_stats['closest_peers']:
                    handles.append(peer.handles[0])
                message.prefix = "%s[%s]" % (message.speaker, "|".join(handles))
            else:
                message.prefix = "%s[%d]" % (message.speaker, len(message_with_stats['closest_peers']))

            self.deliver(message)
            self.state.update_net_chain(message.message_hash)
            message.reporting_peers = message_with_stats['reporting_peers']
            self.rebroadcast(message)

    def peers_present(self):
        present_handles = []
        for handle in self.presence.keys():
            if self.presence[handle] == STATUS_ONLINE or STATUS_AWAY:
                present_handles.append(handle)
        return present_handles

    def report_presence(self):
        # if handle isn't in the presence dict, check if rubbish received and send /join if so
        for handle in self.state.get_peer_handles():
                if self.state.handle_is_online(handle):
                    if self.presence.get(handle) is None:
                        if self.client:
                            self.presence[handle] = STATUS_ONLINE
                            self.client.send_join(handle)
                else:
                    if self.presence.get(handle):
                        if self.client:
                            del self.presence[handle]
                            self.client.send_part(handle)

        # if handle IS in the presence dict, check last message received from handle and send /away
        for handle in self.presence.keys():
            if self.presence[handle] is not STATUS_AWAY:
                if self.state.handle_is_away(handle):
                    if self.client:
                        self.presence[handle] = STATUS_AWAY
                        self.client.send_away(handle)
            # need to check if the handle is back
            else:
                if not self.state.handle_is_away(handle):
                    self.presence[handle] = STATUS_ONLINE
                    self.client.send_back(handle)

    def report_error(self, error_code, message):
        packet_info = message['metadata']["packet_info"]
        address = packet_info[0]
        port = packet_info[1]
        packet_sample = packet_info[2]
        if error_code == STALE_PACKET:
            logging.debug("[%s:%d] -> stale packet: %s" % (address, port, binascii.hexlify(message['message_hash'])))
        elif error_code == DUPLICATE_PACKET:
            logging.debug("[%s:%d] -> duplicate packet: %s" % (address, port, binascii.hexlify(message['message_hash'])))
        elif error_code == MALFORMED_PACKET:
            logging.debug("[%s:%d] -> malformed packet: %s" % (address, port, packet_sample))
        elif error_code == INVALID_SIGNATURE:
            logging.debug("[%s:%d] -> invalid packet signature: %s" % (address, port, packet_sample))
        elif error_code == UNSUPPORTED_VERSION:
            logging.debug("[%s:%d] -> pest version not supported: %s" % (address, port, packet_sample))
        elif error_code == INVALID_HANDLE_ENCODING:
            logging.debug("[%s:%d] -> invalid handle encoding: %s" % (address, port, packet_sample))
        elif error_code == OUT_OF_ORDER_NET:
            self.add_message_to_order_buffer_and_send_getdata(message, ['net_chain'])
        elif error_code == OUT_OF_ORDER_SELF:
            self.add_message_to_order_buffer_and_send_getdata(message, ['self_chain'])
        elif error_code == OUT_OF_ORDER_BOTH:
            self.add_message_to_order_buffer_and_send_getdata(message, ['self_chain', 'net_chain'])

    def add_message_to_order_buffer_and_send_getdata(self, message, broken_chains):
        packet_info = message['metadata']["packet_info"]
        address = packet_info[0]
        port = packet_info[1]
        logging.debug(
            "[%s:%d] -> message received out of order: %s" % (address, port, binascii.hexlify(message['message_hash'])))
        self.order_buffer.add(message)
        for chain in broken_chains:
            if not self.getdata_requests.get(message[chain]):
                GetData(message, chain, self.state).send()
                self.getdata_requests[message[chain]] = time.time()
        self.clean_getdata_requests()

    def clean_getdata_requests(self):
        for message_hash in self.getdata_requests.keys():
            if (self.getdata_requests.get(message_hash) <
                    time.time() - self.state.get_knob('getdata_requests_expiration_seconds')):
                del self.getdata_requests[message_hash]

    def deliver(self, message):
        # it's possible that these messages are from an order buffer
        # dump and their immediate copies may already have been broadcast
        # or vice versa so we need to check the long buffer
        if self.long_buffer.has(message.message_hash):
            return


        # send to the irc client
        if self.client:
            # emit a replay warning if this message is a getdata response and older than the last
            # displayed message
            if message.get_data_response:
                if message.timestamp < self.last_delivered:
                    warning = time.strftime(
                        "Replay: %Y-%m-%d %H:%M:%S:",
                        time.localtime(message.timestamp)
                    )
                    if message.command == BROADCAST:
                        self.client.pest_reply(warning)
                    else:
                        self.client.pest_dm_reply(message.speaker, warning)
            self.client.message_from_station(message)
            self.last_delivered = message.timestamp
        self.long_buffer.intern(message)

    # we only update the address table if the speaker is same as peer
    def conditionally_update_at(self, message, address):
        if message.speaker in message.peer.handles:
            self.state.update_at({
                "handle": message.speaker,
                "address": address[0],
                "port": address[1]
            })

    def rebroadcast(self, message):
        if not message.get_data_response:
            if message.bounces < int(self.state.get_knob("max_bounces")):
                message.bounces = message.bounces + 1
                message.forward()
            else:
                logging.debug("message TTL expired: %s" % message.message_hash)

    def send_rubbish(self):
        if self.client:
            Ignore({
                "speaker": self.client.nickname,
            }, self.state).send()